@@ -82,12 +82,15 @@ import gvar as _gvar
 # add extras and utilities to lsqfit 
 from ._extras import empbayes_fit, wavg 
 from ._utilities import dot as _util_dot
+from ._utilities import _build_chiv_chivw
 from ._utilities import multifit, multiminex, gammaQ
 from ._version import version as __version__
 
 
-_FDATA = collections.namedtuple('_FDATA', ['mean', 'wgt'])
-# Internal data type for nonlinear_fit.unpack_data()
+_FDATA = collections.namedtuple(
+    '_FDATA', 'mean inv_wgts svdcorrection logdet nblocks svdn'
+    )
+# Internal data type for _unpack_data()
 
 class nonlinear_fit(object):
     """ Nonlinear least-squares fit.
@@ -186,35 +189,40 @@ class nonlinear_fit(object):
         extracted from ``prior``. The default value is ``None``; ``p0``
         must be defined if ``prior`` is ``None``.
     :type p0: dictionary, array, string or ``None``
-    :param svdcut: If positive, eigenvalues of the (rescaled) ``y`` 
-        covariance matrix that are smaller than ``svdcut`` times the
-        maximum eigenvalue are replaced by ``svdcut`` times the maximum
-        eigenvalue. If negative, eigenmodes with eigenvalues smaller than
-        ``|svdcut|`` times the largest eigenvalue are discarded. If zero or
-        ``None``, the covariance matrix is left unchanged. If ``svdcut`` is
-        a 2-tuple, the first entry is ``svdcut`` for the ``y`` covariance
-        matrix and the second entry is ``svdcut`` for the prior's
-        covariance matrix.
-    :type svdcut: ``None`` or ``float`` or 2-tuple
-    :param svdnum: If positive, at most ``svdnum`` eigenmodes of the 
-        (rescaled) ``y`` covariance matrix are retained; the modes with the
-        smallest eigenvalues are discarded. ``svdnum`` is ignored if set to
-        ``None``. If ``svdnum`` is a 2-tuple, the first entry is ``svdnum``
-        for the ``y`` covariance matrix and the second entry is ``svdnum``
-        for the prior's covariance matrix.
-    :type svdnum: ``None`` or ``int`` or 2-tuple
+    :param svdcut: If ``svdcut`` is nonzero (not ``None``), *svd* cuts
+        are applied to every block-diagonal sub-matrix of the covariance 
+        matrix for the data ``y`` and ``prior`` (if there is a prior).
+        The blocks are first rescaled so that all 
+        diagonal elements equal 1 -- that is, the blocks are replaced
+        by the correlation matrices for the corresponding subsets of variables. 
+        Then, if ``svdcut > 0``, 
+        eigenvalues of the rescaled matrices that are smaller than ``svdcut`` 
+        times the maximum eigenvalue are replaced by ``svdcut`` times the 
+        maximum eigenvalue. This makes the covariance matrix less singular 
+        and less susceptible to roundoff error. When ``svdcut < 0``, 
+        eigenvalues smaller than ``|svdcut|`` times the maximum eigenvalue 
+        are discarded and the corresponding components in ``y`` and 
+        ``prior`` are zeroed out.
+    :type svdcut: ``None`` or ``float``
     :param debug: Set to ``True`` for extra debugging of the fit function
         and a check for roundoff errors. (Default is ``False``.)
     :type debug: boolean
     :param fitterargs: Dictionary of arguments passed on to 
         :class:`lsqfit.multifit`, which does the fitting.
     """
-    def __init__(self, data=None, fcn=None, prior=None, p0=None, #):
-                svdcut=(1e-15, 1e-15), svdnum=None, debug=False, **kargs): 
+    def __init__(
+        self, data, fcn, prior=None, p0=None, 
+        svdcut=1e-15, debug=False, **kargs
+        ): 
         # capture arguments; initialize parameters 
         self.fitterargs = kargs
-        self.svdcut = svdcut if isinstance(svdcut, tuple) else (svdcut, None)
-        self.svdnum = svdnum if isinstance(svdnum, tuple) else (svdnum, None)
+        if isinstance(svdcut, tuple):
+            svdcut = svdcut[0]
+            warnings.warn(
+                'svdcut = tuple is no longer supported; replace by a single number',
+                UserWarning, stacklevel=2
+                )
+        self.svdcut = svdcut
         self.data = data
         self.p0file = p0 if isinstance(p0, str) else None
         self.p0 = p0 if self.p0file is None else None
@@ -222,27 +230,32 @@ class nonlinear_fit(object):
         self._p = None
         self._palt = None
         self.debug = debug
+        if 'svdnum' in kargs:
+            del kargs['svdnum']
+            warnings.warn(
+                'svdnum is no longer supported by lsqfit', 
+                UserWarning, stacklevel=2
+                )
         cpu_time = time.clock()
         
         # unpack prior,data,fcn,p0 to reconfigure for multifit 
-        prior = _unpack_gvars(prior)
-        if (debug and prior is not None and
-            not all(isinstance(pri, _gvar.GVar) for pri in prior.flat)):
-            raise TypeError("Priors must be GVars.")
-        x, y, prior, fdata = _unpack_data( #
+        x, y, prior, fdata = _unpack_data(
             data=self.data, prior=prior, svdcut=self.svdcut, 
-            svdnum=self.svdnum)  
+            )  
         self.x = x 
         self.y = y   
         self.prior = prior  
-        self.svdcorrection = fdata['svdcorrection']
+        self.svdcorrection = fdata.svdcorrection
+        self.nblocks = fdata.nblocks
+        self.svdn = fdata.svdn
         self.p0 = _unpack_p0(p0=self.p0, p0file=self.p0file, prior=self.prior)
         p0 = self.p0.flatten()  # only need the buffer for multifit 
         flatfcn = _unpack_fcn(fcn=self.fcn, p0=self.p0, y=self.y, x=self.x)
         
         # create fit function chiv for multifit 
-        self._chiv = _build_chiv(fdata=fdata, fcn=flatfcn)
-        self._chivw = self._chiv.chivw
+        self._chiv, self._chivw = _build_chiv_chivw(
+            fdata=fdata, fcn=flatfcn, prior=self.prior
+            )
         self.dof = self._chiv.nf - self.p0.size
         nf = self._chiv.nf
         
@@ -287,21 +300,23 @@ class nonlinear_fit(object):
         self._p = None          # lazy evaluation
         self._transformed_p = None # lazy evaluation
         self._palt = None       # lazy evaluation
-        self.psdev = _reformat(self.p0, [covii**0.5 
-                               for covii in self.cov.diagonal()])
+        self.psdev = _reformat(
+            self.p0, 
+            [covii**0.5 for covii in self.cov.diagonal()]
+            )
         # compute logGBF 
-        if 'logdet_all' not in fdata: 
+        if self.prior is None: 
             self.logGBF = None
         else:
             def logdet(m):
                 (sign, ans) = numpy.linalg.slogdet(m)
                 if sign < 0:
-                    warnings.warn('det(fit.cov) < 0 --- roundoff errors?')
+                    warnings.warn('det(fit.cov) < 0 --- roundoff errors? Try an svd cut.')
                 return ans
                 # return numpy.sum(numpy.log(numpy.linalg.svd(m, compute_uv=False)))
             logdet_cov = logdet(self.cov)
             self.logGBF = 0.5*(
-                logdet_cov - fdata['logdet_all'] - self.chi2 -
+                logdet_cov - fdata.logdet - self.chi2 -
                 self.dof * numpy.log(2. * numpy.pi)
                 )
         
@@ -386,6 +401,18 @@ class nonlinear_fit(object):
 
     def format(self, maxline=0, pstyle='v', nline=None): 
         """ Formats fit output details into a string for printing.
+
+        The output tabulates the ``chi**2`` per degree of freedom of the 
+        fit (``chi2/dof``), the number of degrees of freedom, 
+        the logarithm of the Gaussian Bayes Factor for the fit (``logGBF``),
+        and the number of fit-algorithm iterations needed by the fit. 
+        Optionally, it will also list the best-fit values for the 
+        fit parameters together with the prior for each (in ``[]`` on 
+        each line). It can also list all of the data and the corresponding
+        values from the fit. At the end it lists the *SVD* cut, 
+        the number of eigenmodes modified by the *SVD* cut, the relative 
+        and absolute tolerances used in the fit, and the time in seconds
+        needed to do the fit. 
                         
         :param maxline: Maximum number of data points for which fit 
             results and input data are tabulated. ``maxline<0`` implies
@@ -531,15 +558,12 @@ class nonlinear_fit(object):
             logGBF = '%.5g' % self.logGBF
         except:
             logGBF = str(self.logGBF)
-        if 'all' in self.svdcorrection:
-            descr = " (input data correlated with prior)" 
-        elif 'prior' not in self.svdcorrection:
+        if self.prior is None:
             descr = " (no prior)"
         else:
             descr = ""
         table = ('Least Square Fit%s:\n  chi2/dof [dof] = %.2g [%d]    Q = %s'
-                 '    logGBF = %s' % (descr, chi2_dof, dof, Q, logGBF))
-        table = table+("    itns = %d\n" % self.nit)
+                 '    logGBF = %s\n' % (descr, chi2_dof, dof, Q, logGBF))
         if maxline < 0:
             return table
         
@@ -559,19 +583,16 @@ class nonlinear_fit(object):
             max(w2,10), int(max(w2,10)/2) * ' ', max(w3,10))
         for di,stars in zip(data, collect.stars):
             table += (fst % tuple(di)) + stars + '\n'
-        
-        # tabulate settings 
-        def svdfmt(x):
-            """ Format svd elements. """
-            x1 ,x2 = x
-            x1 = "%.2g" % x1 if x1 is not None else 'None'
-            x2 = "%.2g" % x2 if x2 is not None else 'None'
-            return "(" + x1 + "," + x2 + ")"
-        
-        settings = "\nSettings:\n  svdcut = %s   svdnum = %s" % ( #
-            svdfmt(self.svdcut), svdfmt(self.svdnum))
-        settings += "    reltol/abstol = %.2g/%.2g\n" % ( #
-            self.reltol, self.abstol)
+                
+        settings = "\nSettings:\n  svdcut/n = {svdcut}/{svdn}".format(
+            svdcut=self.svdcut, svdn=self.svdn
+            )
+        settings += "    reltol/abstol = {rel:.2g}/{abs:.2g}".format(
+            rel=self.reltol, abs=self.abstol
+            )
+        settings +="    (itns/time = {itns}/{time:.1f})\n".format(
+            itns=self.nit, time=self.time
+            )
         if pstyle == 'm':
             settings = ""
         if self.alg != "lmsder":
@@ -747,8 +768,8 @@ class nonlinear_fit(object):
         arguments in :class:`lsqfit.nonlinear_fit`.
         """
         pexact = self.pmean if pexact is None else pexact
-        # Note: don't need svdcut/svdnum since these are built into the data_iter
-        fargs = dict(fcn=self.fcn, svdcut=None, svdnum=None, p0=pexact)
+        # Note: don't need svdcut since these are built into the data_iter
+        fargs = dict(fcn=self.fcn, svdcut=None, p0=pexact)
         fargs.update(self.fitterargs)
         fargs.update(kargs)
         for ysim, priorsim in self._simulated_data_iter(
@@ -910,7 +931,7 @@ class transform_p(object):
 
         p["XX"] = exp(p[k])    for k = "log(XX)" or "logXX"
 
-        or
+    or ::
 
         p["XX"] = p[k] ** 2    for k = "sqrt(XX)" or "sqrtXX"
 
@@ -925,27 +946,19 @@ class transform_p(object):
     This is a convenience function. It allows for the 
     rapid replacement of a fit parameter by its 
     logarithm or square root without having to rewrite the
-    fit function --- only the prior need be changed. The
-    decorator needs the keys from the prior, and it needs to be told which
-    argument (numbered from 0) of the fit function is the 
-    parameter dictionary, unless the fit function has 
-    only a single argument::
+    fit function --- only the prior need be changed. The decorator
+    needs to be told if the fit function has an ``x`` as its 
+    first argument, followed by the parameters ``p``::
 
-        @lsqfit.transform_p(prior.keys(), 1)
+        @lsqfit.transform_p(prior.keys(), has_x=True)
         def fitfcn(x, p):
             ...
 
-    or ::
-
-        @lsqfit.transform_p(prior.keys(), 0)
-        def fitfcn(p, other_arg1, ...):
-            ...
-
-    or ::
+    versus ::
 
         @lsqfit.transform_p(prior.keys())
         def fitfcn(p):
-            ...        
+            ...
 
     A list of the specific keys that need transforming can be used instead
     of the list of all keys (``prior.keys()``). The decorator assigns a copy 
@@ -955,20 +968,19 @@ class transform_p(object):
         Other keys can be in ``priorkeys`` provided they do not begin
         with ``'log'`` or ``'sqrt'`` --- they are ignored.
     :type priorkeys: sequence
-    :param pindex: Index of the parameters-dictionary in the argument list 
-        of the fit function. Default value is ``None``; one of 
-        ``pkey`` or ``pindex`` must be specified (i.e., ``not None``),
-        unless the fit function has only a single argument.
-    :type pindex: integer or None
+    :param has_x: Set equal to ``True`` if the fit function is a function
+        of ``x`` and parameters ``p`` (*i.e.*, ``f(x,p)``). Set equal to 
+        ``False`` if the fit function is a function only of the parameters
+        (*i.e.*, ``f(p)``). Default is ``False``.
+    :type pindex: bool
     :param pkey: Name of the parameters-variable in the argument keyword
         dictionary of the fit function. Default value is ``None``; one of 
         ``pkey`` or ``pindex`` must be specified (i.e., ``not None``),
         unless the fit function has only a single argument.
     :type pkey: string or None
     """
-    def __init__(self, priorkeys, pindex=None, pkey=None):
-        self.pindex = pindex
-        self.pkey = pkey
+    def __init__(self, priorkeys, has_x=False):
+        self.pindex = 1 if has_x else 0
         self.log_keys = []
         self.sqrt_keys = []
         self.added_keys = []
@@ -1056,39 +1068,17 @@ class transform_p(object):
 
 
     def __call__(self, f):
-        if self.pindex is None and self.pkey is None:
-            fargs = inspect.getargspec(f)[0]
-            if len(fargs) == 1:
-                self.pindex = 0
-                self.pkey = fargs[0]
-            else:
-                raise ValueError('must specify pindex or pkey')
-        if self.pindex is None:
-            try:
-                self.pindex = inspect.getargspec(f)[0].index(self.pkey)
-            except ValueError:
-                raise ValueError('bad pkey: ' + str(self.pkey))
-        elif self.pkey is None:
-            try:
-                self.pkey = inspect.getargspec(f)[0][self.pindex]
-            except IndexError:
-                raise IndexError('bad pindex: ' + str(self.pindex))
-        else:
-            try:
-                tmp = (self.pkey == inspect.getargspec(f)[0][self.pindex])
-                if not tmp:
-                    raise IndexError
-            except IndexError: 
-                raise ValueError(
-                    "bad pindex, pkey: " + str(self.pindex) +
-                    ', ' + str(self.pkey)
-                    )
+        try:
+            pkey = inspect.getargspec(f)[0][self.pindex]
+        except IndexError:
+            raise IndexError('function has too few arguments')
+        
         @functools.wraps(f)
         def newf(*args, **kargs):
             p = ( 
                 _gvar.BufferDict(args[self.pindex])
                 if self.pindex < len(args) else
-                _gvar.BufferDict(kargs[self.pkey])
+                _gvar.BufferDict(kargs[pkey])
                 )
             for i, j in self.log_keys:
                 p[j] = _gvar.exp(p[i])
@@ -1099,9 +1089,10 @@ class transform_p(object):
                 args[self.pindex] = p
                 args = tuple(args)
             else:
-                kargs[self.pkey] = p
+                kargs[pkey] = p
             # newf.transform_p = self
             return f(*args, **kargs)
+
         newf.transform_p = self
         return newf
 
@@ -1126,7 +1117,7 @@ def _reformat(p, buf):
         ans = numpy.array(buf).reshape(numpy.shape(p))
     return ans
     
-def _unpack_data(data, prior, svdcut, svdnum): 
+def _unpack_data(data, prior, svdcut): 
     """ Unpack data and prior into ``(x, y, prior, fdata)``. 
         
     This routine unpacks ``data`` and ``prior`` into ``x, y, prior, fdata``
@@ -1142,13 +1133,15 @@ def _unpack_data(data, prior, svdcut, svdnum):
     |GVar|\s. In the last case it is assumed that the fit function is a
     function of only the parameters: ``fcn(p)`` --- no ``x``. (This is also
     assumed if ``x = False``.)
-        
-    Output data in ``fdata`` includes: fit decompositions of ``y`` 
-    (``fdata["y"]``) and ``prior`` (``fdata["prior"]``), or of 
-    both together (``fdata["all"]``) if they are correlated. 
-    ``fdata["svdcorrection"]`` contains a dictionary with all *svd* corrections
-    (from ``y`` and ``prior``). ``fdata["logdet_prior"]`` contains
-    the logarithm of the determinant of the prior's covariance matrix.
+    
+    Output data in ``fdata`` is: ``fdata.mean`` containing the mean
+    values of ``y.flat`` and ``prior.flat`` (if there is a prior); 
+    ``fdata.svdcorrection``  containing the *svd* correction to ``y.flat`` 
+    and ``prior.flat``; ``fdata.logdet`` containing the logarithem of the 
+    determinant of the covariance matrix of ``y.flat`` and ``prior.flat``;
+    and ``fdata.inv_wgts`` containing a representation of the inverse
+    of the covariance matrix, after *svd* cuts (see :func:`gvar.svd` 
+    for a description of the format).
     """
     # unpack data tuple 
     data_is_3tuple = False
@@ -1166,47 +1159,26 @@ def _unpack_data(data, prior, svdcut, svdnum):
         y = _unpack_gvars(y)
     else:
         raise ValueError("data tuple wrong length: "+str(len(data)))
+    
+    def _apply_svd(data, svdcut=svdcut):
+        ans, inv_wgts = _gvar.svd(data, svdcut=svdcut, wgts=-1)
+        fdata = _FDATA(
+            mean=_gvar.mean(data.flat), 
+            inv_wgts=inv_wgts,
+            svdcorrection=_gvar.svd.correction,
+            logdet=_gvar.svd.logdet,
+            nblocks=_gvar.svd.nblocks,
+            svdn=_gvar.svd.nmod,
+            )
+        return ans, fdata
 
-    # create svd script 
-    fdata = dict(svdcorrection={})
-        
-    def _apply_svd(k, data, fdata=fdata, svdcut=svdcut, svdnum=svdnum):
-        """ apply svd cut and save related data """
-        i = 1 if k == 'prior' else 0
-        ans = _gvar.svd(data, svdcut=svdcut[i], svdnum=svdnum[i], 
-                        rescale=True, compute_inv=True)
-        fdata[k] = _FDATA(mean=_gvar.mean(data.flat), wgt=_gvar.svd.inv_wgt)
-        fdata['svdcorrection'][k] = _gvar.svd.correction
-        if k == 'prior':
-            fdata['logdet_prior'] = _gvar.svd.logdet
-        elif k == 'y':
-            fdata['logdet_y'] = _gvar.svd.logdet
-        elif k == 'all':
-            fdata['logdet_all'] = _gvar.svd.logdet
-        return ans
-
-    if prior is not None:
-        # have prior 
-        if data_is_3tuple or _gvar.uncorrelated(y.flat, prior.flat):
-            # y uncorrelated with prior 
-            y = _apply_svd('y', y)
-            prior = _apply_svd('prior', prior)
-            fdata['logdet_all'] = fdata['logdet_prior'] + fdata['logdet_y']
-        else:
-            # y correlated with prior 
-            yp = _apply_svd('all', numpy.concatenate((y.flat, prior.flat)))
-            y.flat = yp[:y.size]
-            prior.flat = yp[y.size:]
-            # # compute log(det(cov_pr)) where cov_pr = prior part of cov:
-            # invcov = numpy.sum(numpy.outer(wi, wi) 
-            #                    for wi in _gvar.svd.inv_wgt)
-            # s = _gvar.SVD(invcov[y.size:, y.size:])
-            # # following has minus sign because s is for the inv of cov:
-            # fdata['logdet_prior'] = -numpy.sum(numpy.log(vi) for vi in s.val)
-            # fdata['logdet_y'] = fdata['logdet_all'] - fdata['logdet_prior']
+    if prior is None:
+        y, fdata = _apply_svd(y)
     else:
-        # no prior 
-        y = _apply_svd('y', y)
+        prior = _unpack_gvars(prior)
+        yp, fdata = _apply_svd(numpy.concatenate((y.flat, prior.flat)))
+        y.flat = yp[:y.size]
+        prior.flat = yp[y.size:]
     return x, y, prior, fdata
 
 def _unpack_gvars(g):
@@ -1214,6 +1186,7 @@ def _unpack_gvars(g):
     if g is not None:
         g = _gvar.gvar(g)
         if not hasattr(g, 'flat'):
+            # must be a scalar (ie, not an array and not a dictionary)
             g = numpy.asarray(g)
     return g
 
@@ -1347,64 +1320,6 @@ def _unpack_fcn(fcn, p0, y, x):
                 return yo.flat            
     return nfcn
 
-    
-def _build_chiv(fdata, fcn):
-    """ Build ``chiv`` where ``chi**2=sum(chiv(p)**2)``. """
-    if 'all' in fdata:
-        # y and prior correlated 
-        if fdata['all'].wgt.ndim != 2:
-            raise ValueError("fdata['all'].wgt is 1d! dim = "
-                             +str(fdata['all'].wgt.ndim))
-        def chiv(p, fcn=fcn, fd=fdata['all']):
-            delta = numpy.concatenate((fcn(p), p))-fd.mean
-            return _util_dot(fd.wgt, delta)
-        
-        def chivw(p, fcn=fcn, fd=fdata['all']):
-            delta = numpy.concatenate((fcn(p), p))-fd.mean
-            wgt2 = numpy.sum(numpy.outer(wj, wj) 
-                            for wj in reversed(fd.wgt))
-            return _util_dot(wgt2, delta)
-        chiv.nf = len(fdata['all'].wgt)
-        
-    elif 'prior' in fdata:
-        # y and prior uncorrelated 
-        def chiv(p, fcn=fcn, yfd=fdata['y'], pfd=fdata['prior']):
-            ans = []
-            for d, w in [(fcn(p)-yfd.mean, yfd.wgt), 
-                        (p-pfd.mean, pfd.wgt)]:
-                ans.append(_util_dot(w, d) if w.ndim==2 else w*d)
-            return numpy.concatenate(tuple(ans))
-        def chivw(p, fcn=fcn, yfd=fdata['y'], pfd=fdata['prior']):
-            ans = []
-            for d, w in [(fcn(p)-yfd.mean, yfd.wgt), 
-                        (p-pfd.mean, pfd.wgt)]:
-                if w.ndim == 2:
-                    w2 = numpy.sum(numpy.outer(wj, wj) 
-                                   for wj in reversed(w))
-                    ans.append(_util_dot(w2, d))
-                else:
-                    ans.append(w*w*d)
-            return numpy.concatenate(tuple(ans))
-        chiv.nf = len(fdata['y'].wgt)+len(fdata['prior'].wgt)
-    else:
-        # no prior 
-        def chiv(p, fcn=fcn, fd=fdata['y']):
-            ydelta = fcn(p)-fd.mean
-            return (_util_dot(fd.wgt, ydelta) if fd.wgt.ndim==2 
-                    else fd.wgt*ydelta)
-        def chivw(p, fcn=fcn, fd=fdata['y']):
-            ydelta = fcn(p)-fd.mean
-            if fd.wgt.ndim == 2:
-                wgt2 = numpy.sum(numpy.outer(wj, wj) 
-                                 for wj in reversed(fd.wgt))
-                return _util_dot(wgt2, ydelta)
-            else:
-                return fd.wgt*fd.wgt*ydelta        
-        chiv.nf = len(fdata['y'].wgt)        
-    chiv.chivw = chivw
-    return chiv
-
-
 def _y_fcn_match(y, f):
     if hasattr(f,'keys'):
         f = BufferDict(f)
@@ -1430,11 +1345,11 @@ def _y_fcn_match(y, f):
 
 # legacy definitions (obsolete) 
 class _legacy_constructor:
-    def __init__(self,cl,msg):
+    def __init__(self, cl, msg):
         self.cl = cl
         self.msg = msg
     
-    def __call__(self,*args,**kargs):
+    def __call__(self, *args, **kargs):
         warnings.warn(self.msg, UserWarning, stacklevel=2)
         return self.cl(*args,**kargs)
     
