@@ -458,7 +458,7 @@ class test_gvar2(unittest.TestCase,ArrayTests):
         self.assertEqual(gvar("1e+9 +- 1.23e+6").fmt(), "1.0000(12)e+09")
         self.assertEqual(gvar("1e-9 +- 0").fmt(), "1(0)e-09")
         self.assertEqual(gvar("0(0)").fmt(), "0(0)")
-        self.assertEqual(gvar("1e-9 +- 0.129").fmt(), '1e-09 +- 0.13')
+        self.assertEqual(gvar("1.234e-9 +- 0.129").fmt(), '1e-09 +- 0.13')
         self.assertEqual(gvar("1.23(4)e-9").fmt(), "1.230(40)e-09")
         self.assertEqual(gvar("1.23 +- 1.23e-12").fmt(), "1.23 +- 1.2e-12")
         self.assertEqual(gvar("1.23 +- 1.23e-6").fmt(), "1.2300000(12)")
@@ -679,10 +679,10 @@ class test_gvar2(unittest.TestCase,ArrayTests):
         self.assert_arraysclose(evalcov(z.flat), evalcov(x))
         
     def test_asgvar(self):
-        """ asgvar """
-        z = asgvar(x)
+        """ gvar functions as asgvar """
+        z = gvar(x)
         self.assertTrue(z is x)
-        z = asgvar("2.00(25)")
+        z = gvar("2.00(25)")
         self.assertEqual(z.mean,2.0)
         self.assertEqual(z.sdev,0.25)
     
@@ -828,12 +828,16 @@ class test_gvar2(unittest.TestCase,ArrayTests):
         global x,y,gvar
         n = 1000
         rtol = 5./n**0.5
+        x = gvar(x.mean, x.sdev)
+        y = gvar(y.mean, y.sdev)
         f = raniter([x,y],n)
         ans = [fi for fi in f]
+        # print(x, y, evalcov([x,y]))
+        # print (ans)
         ans = np.array(ans).transpose()
-        self.assert_arraysclose(np.cov(ans[0],ans[1]),evalcov([x,y]),rtol=rtol)
         self.assertAlmostEqual(ans[0].mean(),x.mean,delta=x.sdev*rtol)
         self.assertAlmostEqual(ans[1].mean(),y.mean,delta=y.sdev*rtol)
+        self.assert_arraysclose(np.cov(ans[0],ans[1]),evalcov([x,y]),rtol=rtol)
     
     @unittest.skipIf(FAST,"skipping test_raniter2 for speed")
     def test_raniter2(self):
@@ -859,7 +863,7 @@ class test_gvar2(unittest.TestCase,ArrayTests):
     def test_bootstrap_iter(self):
         """ bootstrap_iter """
         p = BufferDict()
-        p = gv.gvar(1,1)*np.array([1,1])+gv.gvar(0.1,1e-4)*np.array([1,-1])
+        p = gvar(1,1)*np.array([1,1])+gvar(0.1,1e-4)*np.array([1,-1])
         p_sw = np.array([p[0]+p[1],p[0]-p[1]])/2.
         p_cov = evalcov(p_sw.flat)
         p_mean = mean(p_sw.flat)
@@ -964,22 +968,181 @@ class test_gvar2(unittest.TestCase,ArrayTests):
                 m2 = np.sum(np.outer(wj,wj) for wj in s.decomp(2))
                 self.assert_arraysclose(mat,np.dot(m2,minv))
     
+    def test_diagonal_blocks(self):
+        """ find_diagonal_blocks """
+        def make_blocks(*m_list):
+            m_list = [np.asarray(m, float) for m in m_list]
+            n = sum([m.shape[0] for m in m_list])
+            ans = np.zeros((n,n), float)
+            i = 0
+            for m in m_list:
+                j = i + m.shape[0]
+                ans[i:j, i:j] = m 
+                i = j
+            return ans
+        def compare_blocks(b1, b2):
+            s1 = set([tuple(list(b1i)) for b1i in b1])
+            s2 = set([tuple(list(b2i)) for b2i in b2])
+            self.assertEqual(s1, s2)
+        m = make_blocks(
+            [[1]],
+            [[1, 1], [1, 1]], 
+            [[1]]
+            )
+        compare_blocks(find_diagonal_blocks(m), [[0], [3], [1, 2]])
+        m = make_blocks(
+            [[1, 0, 1], [0, 1, 0], [1, 0, 1]],
+            [[1, 1], [1, 1]], 
+            [[1]],
+            [[1]]
+            )
+        compare_blocks(
+            find_diagonal_blocks(m), [[1], [5], [6], [0, 2], [3, 4]]
+            )
+        m = make_blocks(
+            [[1, 0, 1, 1], 
+             [0, 1, 0, 1], 
+             [1, 0, 1, 1],
+             [1, 1, 1, 1]],
+            [[1, 1], [1, 1]], 
+            [[1]],
+            [[1]]
+            )
+        compare_blocks(
+            find_diagonal_blocks(m), [[6], [7], [0, 1, 2, 3] , [4, 5]]
+            )
+
     def test_svd(self):
         """ svd """
-        x = gvar(1,1)
-        dx = gvar(1,1)*1e-2
-        cut = 4e-2
-        for g in [BufferDict({0:x-dx,1:x+dx}),[x-dx,x+dx]]:
-            gm = svd(g,svdcut=cut,rescale=False,compute_inv=True)
-            xm = (gm[1]+gm[0])/2
-            dxm = (gm[1]-gm[0])/2
-            self.assert_gvclose(xm,x)
-            self.assert_gvclose(dxm,gvar(dx.mean,cut**0.5))
-            invcov = sum(np.outer(wi,wi) for wi in svd.inv_wgt)
-            cov = evalcov(gm.flat)
-            self.assert_arraysclose(numpy.dot(cov,invcov),[[1,0],[0,1]])
-            self.assertAlmostEqual(svd.eigen_range,dx.var)
-            self.assertAlmostEqual(numpy.exp(svd.logdet)/dxm.var*xm.var,4.0)
+        def make_mat(wlist, n):
+            ans = np.zeros((n,n), float)
+            i, wgts = wlist[0]
+            if len(i) > 0:
+                ans[i, i] = np.array(wgts) ** 2
+            for i, wgts in wlist[1:]:
+                for w in wgts:
+                    ans[i, i[:, None]] += np.outer(w, w)
+            return ans
+        def test_gvar(a, b):
+            self.assertEqual(a.fmt(4), b.fmt(4))
+        def test_cov(wgts, cov, atol=1e-7):
+            invcov = make_mat(wgts, cov.shape[0])
+            np.testing.assert_allclose(
+                invcov.dot(cov), np.eye(*cov.shape), atol=atol
+                )
+            np.testing.assert_allclose(svd.logdet, np.log(np.linalg.det(cov)))
+        # diagonal
+        f = gvar(['1(2)', '3(4)'])
+        g, wgts = svd(f, svdcut=0.9, wgts=-1)
+        test_gvar(g[0], f[0])
+        test_gvar(g[1], f[1])
+        test_cov(wgts, evalcov(g))
+        self.assertEqual(svd.nmod, 0)
+        self.assertEqual(svd.eigen_range, 1.)
+        
+        # degenerate
+        g, wgts = svd(3 * [gvar('1(1)')], svdcut=1e-10, wgts=-1)
+        test_cov(wgts, evalcov(g), atol=1e-6)
+        self.assertEqual(svd.nmod, 2)
+        self.assertAlmostEqual(svd.eigen_range, 0.0)
+
+        # blocks
+        x = gvar(10 * ['1(1)'])
+        x[:5] += gvar('1(1)')       # half are correlated
+        g = svd(x, svdcut=0.5)
+        self.assertEqual(svd.nmod, 4)
+        p = np.random.permutation(10)
+        gp = svd(x[p], svdcut=0.5)
+        self.assertEqual(svd.nmod, 4)
+        invp = np.argsort(p)
+        np.testing.assert_allclose(evalcov(g), evalcov(gp[invp]))
+        np.testing.assert_allclose(mean(g), mean(gp[invp]))
+
+
+        # cov[i,i] independent of i, cov[i,j] != 0
+        x, dx = gvar(['1(1)', '0.01(1)'])
+        g, wgts = svd([(x+dx)/2, (x-dx)/2.], svdcut=0.2 ** 2, wgts=-1)
+        y = g[0] + g[1] 
+        dy = g[0] - g[1] 
+        test_gvar(y, x)
+        test_gvar(dy, gvar('0.01(20)'))
+        test_cov(wgts, evalcov(g))
+        self.assertEqual(svd.nmod, 1)
+        self.assertAlmostEqual(svd.eigen_range, 0.01**2)
+
+        # negative svdcut
+        x, dx = gvar(['1(1)', '0.01(1)'])
+        g, wgts = svd([(x+dx)/2, (x-dx)/20.], svdcut=-0.2 ** 2, wgts=-1)
+        y = g[0] + g[1] * 10
+        dy = g[0] - g[1] * 10
+        np.testing.assert_allclose(evalcov([y, dy]), [[1, 0], [0, 0]])
+        test_gvar(y, x)
+        test_gvar(dy, gvar('0(0)'))
+        self.assertEqual(svd.dof, 1)
+        self.assertAlmostEqual(svd.eigen_range, 0.01**2)
+
+        # cov[i,i] independent of i, cov[i,j] != 0 --- cut too small
+        x, dx = gvar(['1(1)', '0.01(1)'])
+        g, wgts = svd([(x+dx)/2, (x-dx)/2.], svdcut=0.0099999** 2, wgts=-1)
+        y = g[0] + g[1] 
+        dy = g[0] - g[1] 
+        test_gvar(y, x)
+        test_gvar(dy, dx)
+        test_cov(wgts, evalcov(g))
+        self.assertEqual(svd.nmod, 0)
+        self.assertAlmostEqual(svd.eigen_range, 0.01**2)
+
+
+        # cov[i,i] independent of i after rescaling, cov[i,j] != 0
+        # rescaling turns this into the previous case
+        g, wgts = svd([(x+dx)/2., (x-dx)/20.], svdcut=0.2 ** 2, wgts=-1)
+        y = g[0] + g[1] * 10.
+        dy = g[0] - g[1] * 10.
+        test_gvar(y, x)
+        test_gvar(dy, gvar('0.01(20)'))
+        test_cov(wgts, evalcov(g))
+        self.assertEqual(svd.nmod, 1)
+        self.assertAlmostEqual(svd.eigen_range, 0.01**2)
+
+        # dispersed correlations
+        g2, g4 = gvar(['2(2)', '4(4)'])
+        orig_g = np.array([g2, (x+dx)/2., g4, (x-dx)/20.]) 
+        g, wgts = svd(orig_g, svdcut=0.2 ** 2, wgts=-1)
+        y = g[1] + g[3] * 10.
+        dy = g[1] - g[3] * 10.
+        test_gvar(y, x)
+        test_gvar(dy, gvar('0.01(20)'))
+        test_gvar(g[0], g2)
+        test_gvar(g[2], g4)
+        test_cov(wgts, evalcov(g))
+        self.assertEqual(svd.nmod, 1)
+        self.assertAlmostEqual(svd.eigen_range, 0.01**2)
+        self.assertEqual(svd.nblocks[1], 2)
+        self.assertEqual(svd.nblocks[2], 1)
+
+        # remove svd correction
+        g.flat -= svd.correction
+        y = g[1] + g[3] * 10.
+        dy = g[1] - g[3] * 10.
+        test_gvar(y, x)
+        test_gvar(dy, dx)
+        test_gvar(g[0], g2)
+        test_gvar(g[2], g4)
+        np.testing.assert_allclose(evalcov(g.flat), evalcov(orig_g))
+
+        # bufferdict
+        g = {}
+        g[0] = (x+dx)/2.
+        g[1] = (x-dx)/20.
+        g, wgts = svd({0:(x+dx)/2., 1:(x-dx)/20.}, svdcut=0.2 ** 2, wgts=-1)
+        assert isinstance(g, BufferDict)
+        y = g[0] + g[1] * 10.
+        dy = g[0] - g[1] * 10.
+        test_gvar(y, x)
+        test_gvar(dy, gvar('0.01(20)'))
+        test_cov(wgts, evalcov(g.flat))
+        self.assertEqual(svd.nmod, 1)
+        self.assertAlmostEqual(svd.eigen_range, 0.01**2)
     
     def test_valder(self):
         """ valder_var """
@@ -1022,11 +1185,28 @@ class test_gvar2(unittest.TestCase,ArrayTests):
     
     def test_chi2(self):
         """ chi2(g1, g2) """
+        # uncorrelated
         g = gvar([1., 2.], [1., 2.])
         x = [2., 4.]
         self.assertAlmostEqual(chi2(x,g), 2.)
         self.assertEqual(chi2.dof, 2)
         self.assertAlmostEqual(chi2.Q, 0.36787944, places=2)
+        
+        # correlated
+        g = np.array([g[0]+g[1], g[0]-g[1]])
+        x = np.array([x[0]+x[1], x[0]-x[1]])
+        self.assertAlmostEqual(chi2(x,g), 2.)
+        self.assertEqual(chi2.dof, 2)
+        self.assertAlmostEqual(chi2.Q, 0.36787944, places=2)
+        
+        # correlated with 0 mode and svdcut < 0
+        g = np.array([g[0], g[1], g[0]+g[1]])
+        x = np.array([x[0], x[1], x[0]+x[1]])
+        self.assertAlmostEqual(chi2(x, g, svdcut=-1e-10), 2.)
+        self.assertEqual(chi2.dof, 2)
+        self.assertAlmostEqual(chi2.Q, 0.36787944, places=2)
+        
+        # dictionaries with different keys
         g = dict(a=gvar(1,1), b=[[gvar(2,2)], [gvar(3,3)], [gvar(4,4)]], c=gvar(5,5))
         x = dict(a=2., b=[[4.], [6.]])
         self.assertAlmostEqual(chi2(x,g), 3.)
@@ -1038,6 +1218,8 @@ class test_gvar2(unittest.TestCase,ArrayTests):
         self.assertAlmostEqual(chi2(2., gvar(1,1)), 1.)
         self.assertEqual(chi2.dof, 1)
         self.assertAlmostEqual(chi2.Q, 0.31731051, places=2)
+        
+        # two dictionaries
         g1 = dict(a=gvar(1, 1), b=[gvar(2, 2)])
         g2 = dict(a=gvar(2, 2), b=[gvar(4, 4)])
         self.assertAlmostEqual(chi2(g1, g2), 0.2 + 0.2)
